#!/usr/bin/perl -w

# Perl script for performing various rendering-related tasks
# (c) 2000 by Peter Popov

use File::Find;
use File::Basename;

# Default values of variables used throughout this script
# All of them can be changed through the command line

$Output_Directory  = "./output/";
$Thumbs_Directory  = "./thumbnails/";
$Images_Directory  = "./images/";
$Stream_Directory  = "./streams/";
$Input_Directory   = "./";
@Files_To_Render   = ();
$Input_File        = "";

$Render_Engine     = "megapov";

@Boolean = ("No", "Yes");

$JPEG = 0;
$GIF  = 1;

@Supported_Formats = ();
$Supported_Formats[$JPEG] = "JPEG";
$Supported_Formats[$GIF] = "GIF";

@Filename_Extensions = ();
$Filename_Extensions[$JPEG] = "jpg";
$Filename_Extensions[$GIF] = "gif";


$UNIX    = 0;
$MSWIN32 = 1;
$MACOS   = 2;
$MSDOS   = 3;
$AMIGAOS = 4;
$VMS     = 5;

@Supported_OSs = ();
$Supported_OSs[$UNIX]    = "UNIX";
$Supported_OSs[$MSWIN32] = "MSWin32";
$Supported_OSs[$MACOS]   = "MacOS";
$Supported_OSs[$MSDOS]   = "MSDOS";
$Supported_OSs[$AMIGAOS] = "AmigaOS";
$Supported_OSs[$VMS]     = "VMS";

$Operating_System        = $UNIX;

$Convert_Image           = 1;
$Make_Thumbnail          = 1;
$Thumbnail_Size          = "160x120";
$Image_Format            = $JPEG;
$Jpeg_Quality            = 95;
$Gif_Colors              = 256;
$Jpeg_Thumbnail_Quality  = 50;

$Ini_File                = "./render.ini";
$Image_Magick            = "convert";

# Various options

$Verbose_Level           = 1;    #0 fo silent, 1 for normal, 2 for verbose
$Auto_Continue           = 0;
$Clean                   = 0;
$Newer_Only              = 1;
$Replicate_Tree          = 0;
$Make_Directories        = 0;
$Test_Only               = 0;

# ------------------------------------------------------------------------

# List of command-line options and their indices. KEEP THEM IN SYNC!!!

@Short_Arguments = (
  "-c",             # "--convert"
  "-I",             # "--input-directory"
  "-f",             # "--ini-file"
  "-i",             # "--input-file"
  "-m",             # "--make-thumbnail"
  "-O",             # "--output-directory"
  "-t",             # "--image-format"
  "-gc",            # "--gif-colors"
  "-e",             # "--render-engine"
  "-q",             # "--quiet"
  "-v",             # "--verbose"
  "-nc",            # "--no-conversion"
  "-j",             # "--jpeg-quality"
  "-T",             # "--no-thumbnail"
  "-S",             # "--thumbnail-size"
  "-td",            # "--thumbs-directory"
  "-id",            # "--images-directory"
  "-sd",            # "--stream-directory"
  "-J",             # "--jpeg-thumbnail-quality"
  "-OS",            # "--operating-system"
  "-ac",            # "--auto-continue"
  "-n",             # "--newer-only"
  "-C",             # "--clean"
  "-rt",            # "--replicate-tree"
  "-md",            # "--make-directories"
  "-test"           # "--test-only"
);

@Arguments = (
  "--convert",
  "--input-directory",
  "--ini-file",
  "--input-file",
  "--make-thumbnail",
  "--output-directory",
  "--image-format",
  "--gif-colors",
  "--render-engine",
  "--quiet",
  "--verbose",
  "--no-conversion",
  "--jpeg-quality",
  "--no-thumbnail",
  "--thumbnail-size",
  "--thumbs-directory",
  "--images-directory",
  "--stream-directory",
  "--jpeg-thumbnail-quality",
  "--operating-system",
  "--auto-continue",
  "--newer-only",
  "--clean",
  "--replicate-tree",
  "--make-directories",
  "--test-only"             
);

$CONVERT                = 0;     
$INPUT_DIRECTORY        = 1;     
$INI_FILE               = 2;     
$INPUT_FILE             = 3;     
$MAKE_THUMBNAIL         = 4;     
$OUTPUT_DIRECTORY       = 5;     
$IMAGE_FORMAT           = 6;     
$GIF_COLORS             = 7;     
$RENDER_ENGINE          = 8;     
$QUIET                  = 9;     
$VERBOSE                = 10;     
$NO_CONVERT             = 11;
$JPEG_QUALITY           = 12;
$NO_THUMBNAIL           = 13;
$THUMBNAIL_SIZE         = 14;
$THUMBS_DIRECTORY       = 15;
$IMAGES_DIRECTORY       = 16;
$STREAM_DIRECTORY       = 17;
$JPEG_THUMBNAIL_QUALITY = 18;
$OPERATING_SYSTEM       = 19;
$AUTO_CONTINUE          = 20;
$NEWER_ONLY             = 21;
$CLEAN                  = 22;
$REPLICATE_TREE         = 23;
$MAKE_DIRECTORIES       = 24;
$TEST_ONLY              = 25;

#----------------------------- Subroutines -------------------------------

# ---------
#
# Matches a string to the set of predefined command-line options. Returns
# the token ID of the option that matched. Handles ambiguous or invalid
# otions
#
# Works :)
#
# ---------

sub Parse_Argument
{
  my $Matches = 0;
  
  for (my $j=0; $j < @Arguments; $j++) {
    if ( substr ( $Arguments[$j], 0, length ($_[0]) ) eq $_[0] ) { $Matches++; $Match = $j; }
  }
  
  if ($Matches == 0) {
    for ($j=0; $j < @Short_Arguments; $j++) {
      if ( $Short_Arguments[$j] eq $_[0] ) { $Matches++; $Match = $j; }
    }
  }
  
  if ($Matches == 0) { die ("\nInvalid option ", $_[0], "\n"); }
  if ($Matches > 1) { die ("\nAmbiguous option. '", $_[0], "' is not an unique prefix.\n\n"); }
  return $Match;
}

# ---------
#
# Recurses a directory and returns a list of all .pov files with pathnames.
#
# This subroutine is based on code shamelessly ripped off the manpage :)
#
# Does not explore symbolic links (yet).
#
# ---------

sub Find_All_POV_Files
{
  my $Z=0;
  find sub { $Files_To_Render[$Z++] = $File::Find::name, -d && '/' . "\n" if /\.pov$/i}, @_;
  split (@Files_To_Render, "\n");
}

sub To_Upcase
{
  my $Foo = $_[0];
  $Foo =~ tr/a-z/A-Z/;
  return $Foo;
}

sub To_Lowcase
{
  my $Foo = $_[0];
  $Foo =~ tr/a-z/A-Z/;
  return $Foo;
}

sub Err_No_Value
{
  die "Value expected after ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'')!\n"
}


sub Err_No_Number
{
  die "The ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'') option requires an integer value but ``$_[1]'' doesn't look like one.\n\n";
}
  
sub Err_Out_Of_Range
{
  die "The ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'') option requires a value in the range [$_[2]..$_[3]] but $_[1] is not in it.\n\n";
}

sub Check_File  # (Filename, r, w, x, d)
{
  if (!-e $_[0]) { die "\nFile ``$_[0]'' does not exist.\n";
    if (!($_[1] && -r $_[0])) { die "\nRead permission denied on file ``$_[0]''.\n"; }
    if (!($_[2] && -w $_[0])) { die "\nWrite permission denied on file ``$_[0]''.\n"; }
    if (!($_[3] && -x $_[0])) { die "\nExecute permission denied on file ``$_[0]''.\n"; }
    if (!($_[4] && -d $_[0])) { die "\n``$_[0]'' is not a valid directory.\n"; }
  }
}

sub Display_Extended_Info
{ 
  print "\n", "Current settings:\n";
  print "\n", "Output directory       = $Output_Directory\t For raw ppm output";
  print "\n", "Thumbnails directory   = $Thumbs_Directory\t For thumbnails";
  print "\n", "Images directory       = $Images_Directory\t For full-size compressed images";
  print "\n", "Streams directory      = $Stream_Directory\t For the various streams (debug, render...)";
  print "\n";
  print "\n", "Make dirs if needed    = $Boolean[$Make_Directories]\t Not yet implemented";
  print "\n", "Replicate dir tree     = $Boolean[$Replicate_Tree]\t Not yet implemented";
  print "\n", "Input file             = $Input_File\t The file to render";
  print "\n", "Input directory        = $Input_Directory\t Directory with files for batch render";
  print "\n", "INI file               = $Ini_File\t This file will be used for all files in the batch";
  print "\n", "Auto continue render   = $Boolean[$Auto_Continue]\t Continue render if raw output file found";
  print "\n";
  print "\n", "Convert image          = $Boolean[$Convert_Image]";
  print "\n", "Image format           = $Supported_Formats[$Image_Format]\t GIF and JPEG are currently supported";
  print "\n", "JPEG quality           = $Jpeg_Quality\t [1..100] , only used if JPEG output selected";
  print "\n", "GIF colors             = $Gif_Colors\t [2..256] , only used if GIF output selected)";
  print "\n", "Make thumbnail         = $Boolean[$Make_Thumbnail]";
  print "\n", "Thumbnail size         = $Thumbnail_Size\t Use FOOxBAR format";
  print "\n", "JPEG thumbnail quality = $Jpeg_Thumbnail_Quality\t Only used if JPEG output selected";
  print "\n", "Clean after converting = $Boolean[$Clean]\t Delete raw output file after convertion";
  print "\n", "Newer files only       = $Boolean[$Newer_Only]\t Do not render file if converted image found";
  print "\n";
  print "\n", "Operating system       = $Supported_OSs[$Operating_System]\t OS detection to be automated soon";
  print "\n", "Verbose level          = $Verbose_Level\t It's 2 or higher if you're reading this :)";
  print "\n", "Test mode enabled      = $Boolean[$Test_Only]\t Redirect syscalls to stdout";
  print "\n";
}

sub Help
{ 
  print "Usage: render [option [<value>]]...\n\n";
  print "Use --verbose for some info on the commands and default values\n";
  for ($Z=0; $Z<@Arguments; $Z++) {
    print $Short_Arguments[$Z], "\t", $Arguments[$Z], "\n";
  }
  print "\n";
}

#--------------------------------------------------------------------------

# Parse command line

if (@ARGV) {
MAINLOOP:
  for (my $i=0; $i < @ARGV; $i++) {
    $Current = Parse_Argument ($ARGV[$i]);
    $Another = ($i < @ARGV-1);
    if ($Another) { $Next = $ARGV[$i+1]; }
    
    if ( $Current == $VERBOSE )          { $Verbose_Level = 2;                    next; }
    if ( $Current == $QUIET )            { $Verbose_Level = 0;                    next; }
    if ( $Current == $CONVERT )          { $Convert_Image = 1;                    next; }
    if ( $Current == $NO_CONVERT )       { $Convert_Image = 0;                    next; }
    if ( $Current == $MAKE_THUMBNAIL  )  { $Make_Thumbnail = 1;                   next; }
    if ( $Current == $NO_THUMBNAIL  )    { $Make_Thumbnail = 0;                   next; }
    if ( $Current == $AUTO_CONTINUE )    { $Auto_Continue = 1;                    next; }
    if ( $Current == $REPLICATE_TREE )   { $Replicate_Tree = 1;                   next; }
    if ( $Current == $MAKE_DIRECTORIES ) { $Make_Directories = 1;                 next; }
    if ( $Current == $CLEAN  )           { $Clean = 1;                            next; }
    if ( $Current == $NEWER_ONLY )       { $Newer_Only = 1;                       next; }
    if ( $Current == $TEST_ONLY )        { $Test_Only = 1;                        next; }
                                         
    if ( $Current == $RENDER_ENGINE   )  { if ($Another) {
                                             $Render_Engine = $Next;
                                             Check_File ($Render_Engine,1,0,1,0);
                                           }
                                         }
                                         
    if ( $Current == $INPUT_FILE )       { if ($Another) {
                                             $Input_File = $Next;
                                             Check_File ($Input_File,1,0,0,0);
                                             $Input_Directory = "";
                                             @Files_To_Render=($Input_File);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $GIF_COLORS      )  { if ($Another) {
                                             $Gif_Colors = $Next;
                                             if ($Gif_Colors =~ /^\d+$/) {
                                               if ($Gif_Colors >= 2 && $Gif_Colors <= 256) {
                                                 $i++;  next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Gif_Colors, 2, 256); }
                                             }
                                             else { Err_No_Number ($Current, $Gif_Colors); }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $INPUT_DIRECTORY )  { if ($Another) {
                                             $Input_Directory = $Next;
                                             Check_File ($Input_Directory, 1, 1, 0, 1);
                                             Find_All_POV_Files ($Input_Directory);
                                             $Input_File = "";
                                             $i++;  next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                          
    if ( $Current == $OUTPUT_DIRECTORY ) { if ($Another) {
                                             $Output_Directory = $Next;
                                             Check_File ($Output_Directory, 1, 1, 0, 1);
                                             $i++;  next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                          
    if ( $Current == $JPEG_QUALITY )     { if ($Another) {
                                             $Jpeg_Quality = $Next; 
                                             if ($Jpeg_Quality =~ /^\d+$/) {
                                               if ($Jpeg_Quality >= 0 && $Jpeg_Quality <= 100) {
                                                 $i++; next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Jpeg_Quality, 0, 100) }
                                             }
                                             else { Err_No_Number ($Current, $Jpeg_Quality) }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $JPEG_THUMBNAIL_QUALITY )
                                         { if ($Another) {
                                             $Jpeg_Thumbnail_Quality = $Next; 
                                             if ($Jpeg_Thumbnail_Quality =~ /^\d+$/) {
                                               if ($Jpeg_Thumbnail_Quality >= 0 && $Jpeg_Thumbnail_Quality <= 100) {
                                                 $i++; next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Jpeg_Thumbnail_Quality, 0, 100) }
                                             }
                                             else { Err_No_Number ($Current, $Jpeg_Thumbnail_Quality) }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $IMAGE_FORMAT )     { if ($Another) {
                                             if ($Next eq $JPEG || To_Upcase ($Next) eq "JPEG") {
                                               $Image_Format=$JPEG; $i++; next; }
                                               elsif ($Next eq $GIF || To_Upcase ($Next) eq "GIF") {
                                                 $Image_Format=$GIF; $i++; next; }
                                                 else { die "``$Next'' is not a valid value for the ``$Arguments[$Current]'' option.\n\n"; }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $INI_FILE )         { if ($Another) {
                                             $Ini_File = $Next;
                                             Check_File ($Ini_File, 1, 0, 0, 0);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $THUMBNAIL_SIZE )   { if ($Another) {
                                             $Thumbnail_Size = $Next;
                                             if ($Thumbnail_Size =~ /^\d+x\d+$/) {
                                               $i++; next;
                                             }
                                             else {
                                               die "Invalid value for ``$Arguments[$Current]'' .\n"
                                             }
                                           }
                                         }
                                         
    if ( $Current == $THUMBS_DIRECTORY)  { if ($Another) {
                                             $Thumbs_Directory = $Next;
                                             Check_File($Thumbs_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $IMAGES_DIRECTORY)  { if ($Another) {
                                             $Images_Directory = $Next;
                                             Check_File ($Images_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $STREAM_DIRECTORY)  { if ($Another) {
                                             $Stream_Directory = $Next;
                                             Check_File($Stream_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $OPERATING_SYSTEM ) { if ($Another) {
                                             if (To_Upcase ($Next) eq "VMS") { $Operating_System=$VMS; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "MSDOS") { $Operating_System=$MSDOS; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "MACOS") { $Operating_System=$MACOS; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "MSWIN32") { $Operating_System=$MSWIN32; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "AMIGAOS") { $Operating_System=$AMIGAOS; $i++; next MAINLOOP; }
                                               elsif ((To_Upcase ($Next) eq "UNIX") || (To_Upcase ($Next) eq "LINUX")) { $Operating_System = $UNIX; $i++; next MAINLOOP; }
                                               else { print "``$ARGV[$i+1]'' is not a supported OS, UNIX assumed.\n";
                                               $Operating_System=$UNIX; $i++; next MAINLOOP;
                                             }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }

  }
}
else { Help }

if ($Verbose_Level == 2) { Display_Extended_Info }
    
foreach (@Files_To_Render) {
  ($Stripped_File_Name, $Current_File_Directory, $Suffix) = fileparse($_, '\.[pP][oO][vV]');
  $Suffix =~ //;  # should make -w happy :)
  
  $Render_This_File = 1;
  
  if ( $Newer_Only ) {
    if (-e ($Images_Directory . $Stripped_File_Name . "." . $Filename_Extensions[$Image_Format])) {
      print "File $_ has already been rendered. Skipped.\n";
      $Render_This_File = 0;
    }
  }
  
  if ( $Render_This_File ) {
    $Shellout = 
      $Render_Engine . " " .
      $Ini_File .
      " -i"  . $_ .
      " -o"  . $Output_Directory .
      " -l"  . $Current_File_Directory .
      " +GR" . $Stream_Directory . $Stripped_File_Name . ".render" .
      " +GF" . $Stream_Directory . $Stripped_File_Name . ".fatal" .
      " +GD" . $Stream_Directory . $Stripped_File_Name . ".debug" .
      " +GW" . $Stream_Directory . $Stripped_File_Name . ".warning" .
      " +GS" . $Stream_Directory . $Stripped_File_Name . ".statistics" .
      " 2> " . $Stream_Directory . $Stripped_File_Name . ".status"
      ;
    
    if ($Auto_Continue) {
      if (-e ($Output_Directory . $Stripped_File_Name . ".ppm")) {
        if ($Verbose_Level) {
          print "File ", $Output_Directory . $Stripped_File_Name . ".ppm" .
                " exists, forcing continued render.\n";
        }
      }
    }
    
    if ($Verbose_Level) { print "Now rendering $_... \n"; }
    
     if ($Test_Only) { if ($Verbose_Level == 2) { print $Shellout, "\n"; } }
     else { system ($Shellout); }
     
    if ($Verbose_Level) { print "Done!\n\n"; }
    
      if ( $Convert_Image) {
      $Shellout =
        $Image_Magick .
        " -compress " .
        $Supported_Formats[$Image_Format]
      ;
      if ($Image_Format == $JPEG) {
        $Shellout .= " -quality $Jpeg_Quality ";
      }  
      if ($Image_Format == $GIF) {
        $Shellout .= " -dither " . "-colors $Gif_Colors ";
      }
      $Shellout .= " $Output_Directory" . $Stripped_File_Name . ".ppm ";
      if ($Image_Format == $JPEG) {
        $Shellout .= " $Images_Directory" . $Stripped_File_Name . ".jpg ";
      }
      else {
        $Shellout .= " $Images_Directory" . $Stripped_File_Name . ".gif ";
      }
     
     if ($Verbose_Level) { print "Converting to $Supported_Formats[$Image_Format] format... \n"; }
     
     if ($Test_Only) { if ($Verbose_Level == 2) { print $Shellout, "\n"; } }
     else { system ($Shellout); }
  
     if ($Verbose_Level) { print "Done!\n\n"; }
  
    }
  
    if ($Make_Thumbnail) {
      $Shellout =
        $Image_Magick .
        " -antialias" .
        " -compress " .
        $Supported_Formats[$Image_Format] .
        " -geometry " . $Thumbnail_Size
      ;
      if ($Image_Format == $JPEG) {
        $Shellout .= " -quality $Jpeg_Thumbnail_Quality ";
        $Shellout .= " $Output_Directory" . $Stripped_File_Name . ".ppm ";
        $Shellout .= " $Thumbs_Directory" . "t_" . $Stripped_File_Name . ".jpg";
      }
      else {
        $Shellout .= " -dither " . "-colors $Gif_Colors ";
        $Shellout .= " $Output_Directory" . $Stripped_File_Name . ".ppm ";
        $Shellout .= " $Thumbs_Directory" . "t_" . $Stripped_File_Name . ".gif";
      }
    
     if ($Verbose_Level) { print "Creating a $Thumbnail_Size $Supported_Formats[$Image_Format] thumbnail...\n"; }
  
     if ($Test_Only) { if ($Verbose_Level == 2) {print $Shellout, "\n"; } }
     else { system ($Shellout); }
  
     if ($Verbose_Level) { print "Done!\n\n"; }
  
    }
    
    if ( $Clean) {
      print "Deleting raw output file ...\n";
      $Shellout = "rm " . $Output_Directory . $Stripped_File_Name . ".ppm ";
      if ($Test_Only) { if ($Verbose_Level == 2) {print $Shellout, "\n"; } }
      else { system ($Shellout); }
      print "Done!\n\n";
    }
  }
}